# Copyright (c) 2025 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from mautrix import client as cli
from mautrix.errors import MNotFound
from mautrix.types import EventType, SecretStorageDefaultKeyEventContent

from .key import Key, KeyMetadata
from .types import EncryptedAccountDataEventContent


class Machine:
    client: cli.Client

    def __init__(self, client: cli.Client) -> None:
        self.client = client

    async def get_default_key_id(self) -> str | None:
        try:
            data = await self.client.get_account_data(EventType.SECRET_STORAGE_DEFAULT_KEY)
            return SecretStorageDefaultKeyEventContent.deserialize(data).key
        except (MNotFound, ValueError):
            return None

    async def set_default_key_id(self, key_id: str) -> None:
        await self.client.set_account_data(
            EventType.SECRET_STORAGE_DEFAULT_KEY,
            SecretStorageDefaultKeyEventContent(key=key_id),
        )

    async def get_key_data(self, key_id: str) -> KeyMetadata:
        data = await self.client.get_account_data(f"m.secret_storage.key.{key_id}")
        return KeyMetadata.deserialize(data)

    async def set_key_data(self, key_id: str, data: KeyMetadata) -> None:
        await self.client.set_account_data(f"m.secret_storage.key.{key_id}", data)

    async def get_default_key_data(self) -> tuple[str, KeyMetadata]:
        key_id = await self.get_default_key_id()
        if not key_id:
            raise ValueError("No default key ID set")
        return key_id, await self.get_key_data(key_id)

    async def get_decrypted_account_data(self, event_type: EventType | str, key: Key) -> bytes:
        data = await self.client.get_account_data(event_type)
        parsed = EncryptedAccountDataEventContent.deserialize(data)
        return parsed.decrypt(event_type, key)

    async def set_encrypted_account_data(
        self, event_type: EventType | str, data: bytes, *keys: Key
    ) -> None:
        encrypted_data = {}
        for key in keys:
            encrypted_data[key.id] = key.encrypt(event_type, data)
        await self.client.set_account_data(
            event_type,
            EncryptedAccountDataEventContent(encrypted=encrypted_data),
        )

    async def generate_and_upload_key(self, passphrase: str | None = None) -> Key:
        key = Key.generate(passphrase)
        await self.set_key_data(key.id, key.metadata)
        return key



# Copyright (c) 2025 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import TYPE_CHECKING

from attr import dataclass

from mautrix.types import EventType, SerializableAttrs, SerializableEnum
from mautrix.types.event.account_data import account_data_event_content_map

if TYPE_CHECKING:
    from .key import Key


class Algorithm(SerializableEnum):
    AES_HMAC_SHA2 = "m.secret_storage.v1.aes-hmac-sha2"
    CURVE25519_AES_SHA2 = "m.secret_storage.v1.curve25519-aes-sha2"


class PassphraseAlgorithm(SerializableEnum):
    PBKDF2 = "m.pbkdf2"


@dataclass
class EncryptedKeyData(SerializableAttrs):
    ciphertext: str
    iv: str
    mac: str


@dataclass
class EncryptedAccountDataEventContent(SerializableAttrs):
    encrypted: dict[str, EncryptedKeyData]

    def decrypt(self, event_type: str | EventType, key: "Key") -> bytes:
        try:
            encrypted_data = self.encrypted[key.id]
        except KeyError as e:
            raise ValueError(f"Event not encrypted for provided key") from e
        return key.decrypt(event_type, encrypted_data)


for encrypted_account_data_type in (
    EventType.CROSS_SIGNING_MASTER,
    EventType.CROSS_SIGNING_USER_SIGNING,
    EventType.CROSS_SIGNING_SELF_SIGNING,
    EventType.MEGOLM_BACKUP_V1,
):
    account_data_event_content_map[encrypted_account_data_type] = EncryptedAccountDataEventContent



from .key import Key, KeyMetadata, PassphraseMetadata
from .machine import Machine
from .types import (
    Algorithm,
    EncryptedAccountDataEventContent,
    EncryptedKeyData,
    PassphraseAlgorithm,
)



# Copyright (c) 2025 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from ..types import (
    JSON,
    CrossSigner,
    CrossSigningKeys,
    CrossSigningUsage,
    DeviceIdentity,
    EventType,
    KeyID,
    UserID,
)
from .cross_signing_key import CrossSigningPrivateKeys, CrossSigningPublicKeys, CrossSigningSeeds
from .device_lists import DeviceListMachine
from .signature import sign_olm
from .ssss import Key as SSSSKey


class CrossSigningMachine(DeviceListMachine):
    _cross_signing_public_keys: CrossSigningPublicKeys | None
    _cross_signing_public_keys_fetched: bool
    _cross_signing_private_keys: CrossSigningPrivateKeys | None

    async def verify_with_recovery_key(self, recovery_key: str) -> None:
        if not self.account.shared:
            raise ValueError("Device keys must be shared before verifying with recovery key")
        key_id, key_data = await self.ssss.get_default_key_data()
        ssss_key = key_data.verify_recovery_key(key_id, recovery_key)
        seeds = await self._fetch_cross_signing_keys_from_ssss(ssss_key)
        self._import_cross_signing_keys(seeds)
        await self.sign_own_device(self.own_identity)

    def _import_cross_signing_keys(self, seeds: CrossSigningSeeds) -> None:
        self._cross_signing_private_keys = seeds.to_keys()
        self._cross_signing_public_keys = self._cross_signing_private_keys.public_keys

    async def generate_recovery_key(
        self, passphrase: str | None = None, seeds: CrossSigningSeeds | None = None
    ) -> str:
        if not self.account.shared:
            raise ValueError("Device keys must be shared before generating recovery key")
        seeds = seeds or CrossSigningSeeds.generate()
        ssss_key = await self.ssss.generate_and_upload_key(passphrase)
        await self._upload_cross_signing_keys_to_ssss(ssss_key, seeds)
        await self._publish_cross_signing_keys(seeds.to_keys())
        await self.ssss.set_default_key_id(ssss_key.id)
        await self.sign_own_device(self.own_identity)
        return ssss_key.recovery_key

    async def _fetch_cross_signing_keys_from_ssss(self, key: SSSSKey) -> CrossSigningSeeds:
        return CrossSigningSeeds(
            master_key=await self.ssss.get_decrypted_account_data(
                EventType.CROSS_SIGNING_MASTER, key
            ),
            user_signing_key=await self.ssss.get_decrypted_account_data(
                EventType.CROSS_SIGNING_USER_SIGNING, key
            ),
            self_signing_key=await self.ssss.get_decrypted_account_data(
                EventType.CROSS_SIGNING_SELF_SIGNING, key
            ),
        )

    async def _upload_cross_signing_keys_to_ssss(
        self, key: SSSSKey, seeds: CrossSigningSeeds
    ) -> None:
        await self.ssss.set_encrypted_account_data(
            EventType.CROSS_SIGNING_MASTER, seeds.master_key, key
        )
        await self.ssss.set_encrypted_account_data(
            EventType.CROSS_SIGNING_USER_SIGNING, seeds.user_signing_key, key
        )
        await self.ssss.set_encrypted_account_data(
            EventType.CROSS_SIGNING_SELF_SIGNING, seeds.self_signing_key, key
        )

    async def get_own_cross_signing_public_keys(self) -> CrossSigningPublicKeys | None:
        if self._cross_signing_public_keys or self._cross_signing_public_keys_fetched:
            return self._cross_signing_public_keys
        keys = await self.get_cross_signing_public_keys(self.client.mxid)
        self._cross_signing_public_keys_fetched = True
        if keys:
            self._cross_signing_public_keys = keys
        return keys

    async def get_cross_signing_public_keys(
        self, user_id: UserID
    ) -> CrossSigningPublicKeys | None:
        db_keys = await self.crypto_store.get_cross_signing_keys(user_id)
        if CrossSigningUsage.MASTER not in db_keys and user_id not in self._cs_fetch_attempted:
            self.log.debug(f"Didn't find any cross-signing keys for {user_id}, fetching...")
            async with self._fetch_keys_lock:
                if user_id not in self._cs_fetch_attempted:
                    self._cs_fetch_attempted.add(user_id)
                    await self._fetch_keys([user_id], include_untracked=True)
            db_keys = await self.crypto_store.get_cross_signing_keys(user_id)
        if CrossSigningUsage.MASTER not in db_keys:
            return None
        return CrossSigningPublicKeys(
            master_key=db_keys[CrossSigningUsage.MASTER].key,
            self_signing_key=(
                db_keys[CrossSigningUsage.SELF].key if CrossSigningUsage.SELF in db_keys else None
            ),
            user_signing_key=(
                db_keys[CrossSigningUsage.USER].key if CrossSigningUsage.USER in db_keys else None
            ),
        )

    async def sign_own_device(self, device: DeviceIdentity) -> None:
        full_keys = await self._get_full_device_keys(device)
        ssk = self._cross_signing_private_keys.self_signing_key
        signature = sign_olm(full_keys, ssk)
        full_keys.signatures = {self.client.mxid: {KeyID.ed25519(ssk.public_key): signature}}
        await self.client.upload_one_signature(device.user_id, device.device_id, full_keys)
        await self.crypto_store.put_signature(
            CrossSigner(device.user_id, device.signing_key),
            CrossSigner(self.client.mxid, ssk.public_key),
            signature,
        )

    async def _publish_cross_signing_keys(
        self,
        keys: CrossSigningPrivateKeys,
        auth: dict[str, JSON] | None = None,
    ) -> None:
        public = keys.public_keys
        master_key = CrossSigningKeys(
            user_id=self.client.mxid,
            usage=[CrossSigningUsage.MASTER],
            keys={KeyID.ed25519(public.master_key): public.master_key},
        )
        master_key.signatures = {
            self.client.mxid: {
                KeyID.ed25519(self.client.device_id): sign_olm(master_key, self.account),
            }
        }
        self_key = CrossSigningKeys(
            user_id=self.client.mxid,
            usage=[CrossSigningUsage.SELF],
            keys={KeyID.ed25519(public.self_signing_key): public.self_signing_key},
        )
        self_key.signatures = {
            self.client.mxid: {
                KeyID.ed25519(public.master_key): sign_olm(self_key, keys.master_key),
            }
        }
        user_key = CrossSigningKeys(
            user_id=self.client.mxid,
            usage=[CrossSigningUsage.USER],
            keys={KeyID.ed25519(public.user_signing_key): public.user_signing_key},
        )
        user_key.signatures = {
            self.client.mxid: {
                KeyID.ed25519(public.master_key): sign_olm(user_key, keys.master_key),
            }
        }
        await self.client.upload_cross_signing_keys(
            keys={
                CrossSigningUsage.MASTER: master_key,
                CrossSigningUsage.SELF: self_key,
                CrossSigningUsage.USER: user_key,
            },
            auth=auth,
        )
        await self.crypto_store.put_cross_signing_key(
            self.client.mxid, CrossSigningUsage.MASTER, public.master_key
        )
        await self.crypto_store.put_cross_signing_key(
            self.client.mxid, CrossSigningUsage.SELF, public.self_signing_key
        )
        await self.crypto_store.put_cross_signing_key(
            self.client.mxid, CrossSigningUsage.USER, public.user_signing_key
        )
        self._cross_signing_private_keys = keys
        self._cross_signing_public_keys = public



# Copyright (c) 2025 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import hashlib
import hmac

import base58
import unpaddedbase64

from mautrix.types import EventType

try:
    from Crypto import Random
    from Crypto.Cipher import AES
    from Crypto.Hash import SHA256
    from Crypto.Protocol.KDF import HKDF
    from Crypto.Util import Counter
except ImportError:
    from Cryptodome import Random
    from Cryptodome.Cipher import AES
    from Cryptodome.Hash import SHA256
    from Cryptodome.Protocol.KDF import HKDF
    from Cryptodome.Util import Counter

cryptorand = Random.new()


def decode_base58_recovery_key(key: str) -> bytes | None:
    key_bytes = base58.b58decode(key.replace(" ", ""))
    if len(key_bytes) != 35 or key_bytes[0] != 0x8B or key_bytes[1] != 1:
        return None
    parity = 0
    for byte in key_bytes[:34]:
        parity ^= byte
    return key_bytes[2:34] if parity == key_bytes[34] else None


def encode_base58_recovery_key(key: bytes) -> str:
    key_bytes = bytearray(35)
    key_bytes[0] = 0x8B
    key_bytes[1] = 1
    key_bytes[2:34] = key
    parity = 0
    for byte in key_bytes:
        parity ^= byte
    key_bytes[34] = parity
    encoded_key = base58.b58encode(key_bytes).decode("utf-8")
    return " ".join(encoded_key[i : i + 4] for i in range(0, len(encoded_key), 4))


def derive_keys(key: bytes, name: str | EventType = "") -> tuple[bytes, bytes]:
    aes_key, hmac_key = HKDF(
        master=key,
        key_len=32,
        salt=b"\x00" * 32,
        hashmod=SHA256,
        num_keys=2,
        context=str(name).encode("utf-8"),
    )
    return aes_key, hmac_key


def prepare_aes(key: bytes, iv: str | bytes) -> AES:
    if isinstance(iv, str):
        iv = unpaddedbase64.decode_base64(iv)
    # initial_value = struct.unpack(">Q", iv[8:])[0]
    # counter = Counter.new(64, prefix=iv[:8], initial_value=initial_value)
    counter = Counter.new(128, initial_value=int.from_bytes(iv, byteorder="big"))
    return AES.new(key=key, mode=AES.MODE_CTR, counter=counter)


def calculate_hash(key: bytes, iv: str | bytes) -> str:
    aes_key, hmac_key = derive_keys(key)
    cipher = prepare_aes(aes_key, iv).decrypt(b"\x00" * 32)
    digest = hmac.digest(hmac_key, cipher, hashlib.sha256)
    return unpaddedbase64.encode_base64(digest)



# Copyright (c) 2025 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, TypedDict
import functools
import json

import olm
import unpaddedbase64

from mautrix.types import (
    JSON,
    DeviceID,
    EncryptionKeyAlgorithm,
    KeyID,
    Serializable,
    Signature,
    SigningKey,
    UserID,
)

try:
    from Crypto.PublicKey import ECC
    from Crypto.Signature import eddsa
except ImportError:
    from Cryptodome.PublicKey import ECC
    from Cryptodome.Signature import eddsa

canonical_json = functools.partial(
    json.dumps, ensure_ascii=False, separators=(",", ":"), sort_keys=True
)


class SignedObject(TypedDict):
    signatures: dict[UserID, dict[str, str]]
    unsigned: Any


def sign_olm(data: dict[str, JSON] | Serializable, key: olm.PkSigning | olm.Account) -> Signature:
    if isinstance(data, Serializable):
        data = data.serialize()
    data.pop("signatures", None)
    data.pop("unsigned", None)
    return Signature(key.sign(canonical_json(data)))


def verify_signature_json(
    data: "SignedObject", user_id: UserID, key_name: DeviceID | str, key: SigningKey
) -> bool:
    data_copy = {**data}
    data_copy.pop("unsigned", None)
    signatures = data_copy.pop("signatures")
    key_id = str(KeyID(EncryptionKeyAlgorithm.ED25519, key_name))
    try:
        signature = signatures[user_id][key_id]
        decoded_key = unpaddedbase64.decode_base64(key)
        # pycryptodome doesn't accept raw keys, so wrap it in a DER structure
        der_key = b"\x30\x2a\x30\x05\x06\x03\x2b\x65\x70\x03\x21\x00" + decoded_key
        decoded_signature = unpaddedbase64.decode_base64(signature)
        parsed_key = ECC.import_key(der_key)
        verifier = eddsa.new(parsed_key, "rfc8032")
        verifier.verify(canonical_json(data_copy).encode("utf-8"), decoded_signature)
        return True
    except (KeyError, ValueError):
        return False



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, AsyncContextManager, Type
from abc import ABC, abstractmethod
from contextlib import asynccontextmanager
from contextvars import ContextVar
import logging

from yarl import URL

from mautrix import __optional_imports__
from mautrix.util.logging import TraceLogger

from .connection import LoggingConnection
from .errors import DatabaseNotOwned, ForeignTablesFound
from .scheme import Scheme
from .upgrade import UpgradeTable, upgrade_tables

if __optional_imports__:
    from aiosqlite import Cursor
    from asyncpg import Record

conn_var: ContextVar[LoggingConnection | None] = ContextVar("db_connection", default=None)


class Database(ABC):
    schemes: dict[str, Type[Database]] = {}
    log: TraceLogger

    scheme: Scheme
    url: URL
    _db_args: dict[str, Any]
    upgrade_table: UpgradeTable | None
    owner_name: str | None
    ignore_foreign_tables: bool

    def __init__(
        self,
        url: URL,
        upgrade_table: UpgradeTable | None,
        db_args: dict[str, Any] | None = None,
        log: TraceLogger | None = None,
        owner_name: str | None = None,
        ignore_foreign_tables: bool = True,
    ) -> None:
        self.url = url
        self._db_args = {**db_args} if db_args else {}
        self.upgrade_table = upgrade_table
        self.owner_name = owner_name
        self.ignore_foreign_tables = ignore_foreign_tables
        self.log = log or logging.getLogger("mau.db")
        assert isinstance(self.log, TraceLogger)

    @classmethod
    def create(
        cls,
        url: str | URL,
        *,
        db_args: dict[str, Any] | None = None,
        upgrade_table: UpgradeTable | str | None = None,
        log: logging.Logger | TraceLogger | None = None,
        owner_name: str | None = None,
        ignore_foreign_tables: bool = True,
    ) -> Database:
        url = URL(url)
        try:
            impl = cls.schemes[url.scheme]
        except KeyError as e:
            if url.scheme in ("postgres", "postgresql"):
                raise RuntimeError(
                    f"Unknown database scheme {url.scheme}."
                    " Perhaps you forgot to install asyncpg?"
                ) from e
            elif url.scheme in ("sqlite", "sqlite3"):
                raise RuntimeError(
                    f"Unknown database scheme {url.scheme}."
                    " Perhaps you forgot to install aiosqlite?"
                ) from e
            raise RuntimeError(f"Unknown database scheme {url.scheme}") from e
        if isinstance(upgrade_table, str):
            upgrade_table = upgrade_tables[upgrade_table]
        elif upgrade_table is None:
            upgrade_table = UpgradeTable()
        elif not isinstance(upgrade_table, UpgradeTable):
            raise ValueError(f"Can't use {type(upgrade_table)} as the upgrade table")
        return impl(
            url,
            db_args=db_args,
            upgrade_table=upgrade_table,
            log=log,
            owner_name=owner_name,
            ignore_foreign_tables=ignore_foreign_tables,
        )

    def override_pool(self, db: Database) -> None:
        pass

    async def start(self) -> None:
        if not self.ignore_foreign_tables:
            await self._check_foreign_tables()
        if self.owner_name:
            await self._check_owner()
        if self.upgrade_table and len(self.upgrade_table.upgrades) > 0:
            await self.upgrade_table.upgrade(self)

    async def _check_foreign_tables(self) -> None:
        if await self.table_exists("state_groups_state"):
            raise ForeignTablesFound("found state_groups_state likely belonging to Synapse")
        elif await self.table_exists("roomserver_rooms"):
            raise ForeignTablesFound("found roomserver_rooms likely belonging to Dendrite")

    async def _check_owner(self) -> None:
        await self.execute(
            """CREATE TABLE IF NOT EXISTS database_owner (
                key   INTEGER PRIMARY KEY DEFAULT 0,
                owner TEXT NOT NULL
            )"""
        )
        owner = await self.fetchval("SELECT owner FROM database_owner WHERE key=0")
        if not owner:
            await self.execute("INSERT INTO database_owner (owner) VALUES ($1)", self.owner_name)
        elif owner != self.owner_name:
            raise DatabaseNotOwned(owner)

    @abstractmethod
    async def stop(self) -> None:
        pass

    @abstractmethod
    def acquire_direct(self) -> AsyncContextManager[LoggingConnection]:
        pass

    @asynccontextmanager
    async def acquire(self) -> LoggingConnection:
        conn = conn_var.get(None)
        if conn is not None:
            yield conn
            return
        async with self.acquire_direct() as conn:
            token = conn_var.set(conn)
            try:
                yield conn
            finally:
                conn_var.reset(token)

    async def execute(self, query: str, *args: Any, timeout: float | None = None) -> str | Cursor:
        async with self.acquire() as conn:
            return await conn.execute(query, *args, timeout=timeout)

    async def executemany(
        self, query: str, *args: Any, timeout: float | None = None
    ) -> str | Cursor:
        async with self.acquire() as conn:
            return await conn.executemany(query, *args, timeout=timeout)

    async def fetch(self, query: str, *args: Any, timeout: float | None = None) -> list[Record]:
        async with self.acquire() as conn:
            return await conn.fetch(query, *args, timeout=timeout)

    async def fetchval(
        self, query: str, *args: Any, column: int = 0, timeout: float | None = None
    ) -> Any:
        async with self.acquire() as conn:
            return await conn.fetchval(query, *args, column=column, timeout=timeout)

    async def fetchrow(
        self, query: str, *args: Any, timeout: float | None = None
    ) -> Record | None:
        async with self.acquire() as conn:
            return await conn.fetchrow(query, *args, timeout=timeout)

    async def table_exists(self, name: str) -> bool:
        async with self.acquire() as conn:
            return await conn.table_exists(name)



# Copyright (c) 2025 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import NamedTuple

import olm

from mautrix.crypto.ssss.util import cryptorand
from mautrix.types import SigningKey


class CrossSigningPublicKeys(NamedTuple):
    master_key: SigningKey
    self_signing_key: SigningKey
    user_signing_key: SigningKey


class CrossSigningPrivateKeys(NamedTuple):
    master_key: olm.PkSigning
    self_signing_key: olm.PkSigning
    user_signing_key: olm.PkSigning

    @property
    def public_keys(self) -> CrossSigningPublicKeys:
        return CrossSigningPublicKeys(
            master_key=self.master_key.public_key,
            self_signing_key=self.self_signing_key.public_key,
            user_signing_key=self.user_signing_key.public_key,
        )


class CrossSigningSeeds(NamedTuple):
    master_key: bytes
    self_signing_key: bytes
    user_signing_key: bytes

    def to_keys(self) -> CrossSigningPrivateKeys:
        return CrossSigningPrivateKeys(
            master_key=olm.PkSigning(self.master_key),
            self_signing_key=olm.PkSigning(self.self_signing_key),
            user_signing_key=olm.PkSigning(self.user_signing_key),
        )

    @classmethod
    def generate(cls) -> "CrossSigningSeeds":
        return cls(
            master_key=cryptorand.read(32),
            self_signing_key=cryptorand.read(32),
            user_signing_key=cryptorand.read(32),
        )



# Copyright (c) 2025 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from mautrix.types import SigningKey, UserID

from .signature import verify_signature_json


def test_verify_signature_json() -> None:
    assert verify_signature_json(
        # This is actually a federation PDU rather than a device signature,
        # but they're both 25519 curves so it doesn't make a difference.
        {
            "auth_events": [
                "$L8Ak6A939llTRIsZrytMlLDXQhI4uLEjx-wb1zSg-Bw",
                "$QJmr7mmGeXGD4Tof0ZYSPW2oRGklseyHTKtZXnF-YNM",
                "$7bkKK_Z-cGQ6Ae4HXWGBwXyZi3YjC6rIcQzGfVyl3Eo",
            ],
            "content": {},
            "depth": 3212,
            "hashes": {"sha256": "K549YdTnv62Jn84Y7sS5ZN3+AdmhleZHbenbhUpR2R8"},
            "origin_server_ts": 1754242687127,
            "prev_events": ["$DAhJg4jVsqk5FRatE2hbT1dSA8D2ASy5DbjEHIMSHwY"],
            "room_id": "!offtopic-2:continuwuity.org",
            "sender": "@tulir:maunium.net",
            "type": "m.room.message",
            "signatures": {
                UserID("maunium.net"): {
                    "ed25519:a_xxeS": "SkzZdZ+rH22kzCBBIAErTdB0Vg6vkFmzvwjlOarGul72EnufgtE/tJcd3a8szAdK7f1ZovRyQxDgVm/Ib2u0Aw"
                }
            },
            "unsigned": {"age_ts": 1754242687146},
        },
        UserID("maunium.net"),
        "a_xxeS",
        SigningKey("lVt/CC3tv74OH6xTph2JrUmeRj/j+1q0HVa0Xf4QlCg"),
    )



# Copyright (c) 2025 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import pytest

from ...types.event.type import EventType
from .key import Key, KeyMetadata
from .types import EncryptedAccountDataEventContent

KEY1_CROSS_SIGNING_MASTER_KEY = """{
  "encrypted": {
    "gEJqbfSEMnP5JXXcukpXEX1l0aI3MDs0": {
      "iv": "BpKP9nQJTE9jrsAssoxPqQ==",
      "ciphertext": "fNRiiiidezjerTgV+G6pUtmeF3izzj5re/mVvY0hO2kM6kYGrxLuIu2ej80=",
      "mac": "/gWGDGMyOLmbJp+aoSLh5JxCs0AdS6nAhjzpe+9G2Q0="
    }
  }
}"""

KEY1_CROSS_SIGNING_MASTER_KEY_DECRYPTED = bytes(
    [
        0x68,
        0xF9,
        0x7F,
        0xD1,
        0x92,
        0x2E,
        0xEC,
        0xF6,
        0xB8,
        0x2B,
        0xB8,
        0x90,
        0xD2,
        0x4D,
        0x06,
        0x52,
        0x98,
        0x4E,
        0x7A,
        0x1D,
        0x70,
        0x3B,
        0x9E,
        0x86,
        0x7B,
        0x7E,
        0xBA,
        0xF7,
        0xFE,
        0xB9,
        0x5B,
        0x6F,
    ]
)

KEY1_META = """{
  "algorithm": "m.secret_storage.v1.aes-hmac-sha2",
  "passphrase": {
    "algorithm": "m.pbkdf2",
    "iterations": 500000,
    "salt": "y863BOoqOadgDp8S3FtHXikDJEalsQ7d"
  },
  "iv": "xxkTK0L4UzxgAFkQ6XPwsw",
  "mac": "MEhooO0ZhFJNxUhvRMSxBnJfL20wkLgle3ocY0ee/eA"
}"""
KEY1_ID = "gEJqbfSEMnP5JXXcukpXEX1l0aI3MDs0"
KEY1_RECOVERY_KEY = "EsTE s92N EtaX s2h6 VQYF 9Kao tHYL mkyL GKMh isZb KJ4E tvoC"
KEY1_PASSPHRASE = "correct horse battery staple"

KEY2_META = """{
  "algorithm": "m.secret_storage.v1.aes-hmac-sha2",
  "iv": "O0BOvTqiIAYjC+RMcyHfWw==",
  "mac": "7k6OruQlWg0UmQjxGZ0ad4Q6DdwkgnoI7G6X3IjBYtI="
}"""
KEY2_ID = "NVe5vK6lZS9gEMQLJw0yqkzmE5Mr7dLv"
KEY2_RECOVERY_KEY = "EsUC xSxt XJgQ dz19 8WBZ rHdE GZo7 ybsn EFmG Y5HY MDAG GNWe"

KEY2_META_BROKEN_IV = """{
  "algorithm": "m.secret_storage.v1.aes-hmac-sha2",
  "iv": "O0BOvTqiIAYjC+RMcyHfWwMeowMeowMeow",
  "mac": "7k6OruQlWg0UmQjxGZ0ad4Q6DdwkgnoI7G6X3IjBYtI="
}"""

KEY2_META_BROKEN_MAC = """{
  "algorithm": "m.secret_storage.v1.aes-hmac-sha2",
  "iv": "O0BOvTqiIAYjC+RMcyHfWw==",
  "mac": "7k6OruQlWg0UmQjxGZ0ad4Q6DdwkgnoI7G6X3IjBYtIMeowMeowMeow"
}"""


def get_key_meta(meta: str) -> KeyMetadata:
    return KeyMetadata.parse_json(meta)


def get_key1() -> Key:
    return get_key_meta(KEY1_META).verify_recovery_key(KEY1_ID, KEY1_RECOVERY_KEY)


def get_key2() -> Key:
    return get_key_meta(KEY2_META).verify_recovery_key(KEY2_ID, KEY2_RECOVERY_KEY)


def get_encrypted_master_key() -> EncryptedAccountDataEventContent:
    return EncryptedAccountDataEventContent.parse_json(KEY1_CROSS_SIGNING_MASTER_KEY)


def test_decrypt_success() -> None:
    key = get_key1()
    emk = get_encrypted_master_key()
    assert (
        emk.decrypt(EventType.CROSS_SIGNING_MASTER, key) == KEY1_CROSS_SIGNING_MASTER_KEY_DECRYPTED
    )


def test_decrypt_fail_wrong_key() -> None:
    key = get_key2()
    emk = get_encrypted_master_key()
    with pytest.raises(ValueError):
        emk.decrypt(EventType.CROSS_SIGNING_MASTER, key)


def test_decrypt_fail_fake_key() -> None:
    key = get_key2()
    key.id = KEY1_ID
    emk = get_encrypted_master_key()
    with pytest.raises(ValueError):
        emk.decrypt(EventType.CROSS_SIGNING_MASTER, key)


def test_decrypt_fail_wrong_type() -> None:
    key = get_key1()
    emk = get_encrypted_master_key()
    with pytest.raises(ValueError):
        emk.decrypt(EventType.CROSS_SIGNING_SELF_SIGNING, key)


def test_encrypt_roundtrip() -> None:
    key = get_key1()
    data = bytes([0xDE, 0xAD, 0xBE, 0xEF])
    ciphertext = key.encrypt("net.maunium.data", data)
    plaintext = key.decrypt("net.maunium.data", ciphertext)
    assert plaintext == data


def test_verify_recovery_key_correct() -> None:
    meta = get_key_meta(KEY1_META)
    key = meta.verify_recovery_key(KEY1_ID, KEY1_RECOVERY_KEY)
    assert key.recovery_key == KEY1_RECOVERY_KEY


def test_verify_recovery_key_correct2() -> None:
    meta = get_key_meta(KEY2_META)
    key = meta.verify_recovery_key(KEY2_ID, KEY2_RECOVERY_KEY)
    assert key.recovery_key == KEY2_RECOVERY_KEY


def test_verify_recovery_key_invalid() -> None:
    meta = get_key_meta(KEY1_META)
    with pytest.raises(ValueError):
        meta.verify_recovery_key(KEY1_ID, "foo")


def test_verify_recovery_key_incorrect() -> None:
    meta = get_key_meta(KEY1_META)
    with pytest.raises(ValueError):
        meta.verify_recovery_key(KEY2_ID, KEY2_RECOVERY_KEY)


def test_verify_recovery_key_broken_iv() -> None:
    meta = get_key_meta(KEY2_META_BROKEN_IV)
    with pytest.raises(ValueError):
        meta.verify_recovery_key(KEY2_ID, KEY2_RECOVERY_KEY)


def test_verify_recovery_key_broken_mac() -> None:
    meta = get_key_meta(KEY2_META_BROKEN_MAC)
    with pytest.raises(ValueError):
        meta.verify_recovery_key(KEY2_ID, KEY2_RECOVERY_KEY)


def test_verify_passphrase_correct() -> None:
    meta = get_key_meta(KEY1_META)
    key = meta.verify_passphrase(KEY1_ID, KEY1_PASSPHRASE)
    assert key.recovery_key == KEY1_RECOVERY_KEY


def test_verify_passphrase_incorrect() -> None:
    meta = get_key_meta(KEY1_META)
    with pytest.raises(ValueError):
        meta.verify_passphrase(KEY1_ID, "incorrect horse battery staple")


def test_verify_passphrase_notset() -> None:
    meta = get_key_meta(KEY2_META)
    with pytest.raises(ValueError):
        meta.verify_passphrase(KEY2_ID, "hmm")



# Copyright (c) 2025 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional
import base64
import hashlib
import hmac

from attr import dataclass
import unpaddedbase64

from mautrix.types import EventType, SerializableAttrs

from .types import Algorithm, EncryptedKeyData, PassphraseAlgorithm
from .util import (
    calculate_hash,
    cryptorand,
    decode_base58_recovery_key,
    derive_keys,
    encode_base58_recovery_key,
    prepare_aes,
)

try:
    from Crypto.Cipher import AES
    from Crypto.Util import Counter
except ImportError:
    from Cryptodome.Cipher import AES
    from Cryptodome.Util import Counter


@dataclass
class PassphraseMetadata(SerializableAttrs):
    algorithm: PassphraseAlgorithm
    iterations: int
    salt: str
    bits: int = 256

    def get_key(self, passphrase: str) -> bytes:
        if self.algorithm != PassphraseAlgorithm.PBKDF2:
            raise ValueError(f"Unsupported passphrase algorithm {self.algorithm}")
        return hashlib.pbkdf2_hmac(
            "sha512",
            passphrase.encode("utf-8"),
            self.salt.encode("utf-8"),
            self.iterations,
            self.bits // 8,
        )


@dataclass
class KeyMetadata(SerializableAttrs):
    algorithm: Algorithm

    iv: str | None = None
    mac: str | None = None

    name: str | None = None
    passphrase: Optional[PassphraseMetadata] = None

    def verify_passphrase(self, key_id: str, phrase: str) -> "Key":
        if not self.passphrase:
            raise ValueError("Passphrase not set on this key")
        return self.verify_raw_key(key_id, self.passphrase.get_key(phrase))

    def verify_recovery_key(self, key_id: str, recovery_key: str) -> "Key":
        decoded_key = decode_base58_recovery_key(recovery_key)
        if not decoded_key:
            raise ValueError("Invalid recovery key syntax")
        return self.verify_raw_key(key_id, decoded_key)

    def verify_raw_key(self, key_id: str, key: bytes) -> "Key":
        if self.mac.rstrip("=") != calculate_hash(key, self.iv):
            raise ValueError("Key MAC does not match")
        return Key(id=key_id, key=key, metadata=self)


@dataclass
class Key:
    id: str
    key: bytes
    metadata: KeyMetadata

    @classmethod
    def generate(cls, passphrase: str | None = None) -> "Key":
        passphrase_meta = (
            PassphraseMetadata(
                algorithm=PassphraseAlgorithm.PBKDF2,
                iterations=500_000,
                salt=base64.b64encode(cryptorand.read(24)).decode("utf-8"),
                bits=256,
            )
            if passphrase
            else None
        )
        key = passphrase_meta.get_key(passphrase) if passphrase else cryptorand.read(32)
        iv = unpaddedbase64.encode_base64(cryptorand.read(16))
        metadata = KeyMetadata(
            algorithm=Algorithm.AES_HMAC_SHA2,
            passphrase=passphrase_meta,
            mac=calculate_hash(key, iv),
            iv=iv,
        )
        key_id = unpaddedbase64.encode_base64(cryptorand.read(24))
        return cls(key=key, id=key_id, metadata=metadata)

    @property
    def recovery_key(self) -> str:
        return encode_base58_recovery_key(self.key)

    def encrypt(self, event_type: str | EventType, data: str | bytes) -> EncryptedKeyData:
        if isinstance(data, str):
            data = data.encode("utf-8")
        data = base64.b64encode(data).rstrip(b"=")

        aes_key, hmac_key = derive_keys(self.key, event_type)
        iv = bytearray(cryptorand.read(16))
        iv[8] &= 0x7F
        ciphertext = prepare_aes(aes_key, iv).encrypt(data)
        digest = hmac.digest(hmac_key, ciphertext, hashlib.sha256)
        return EncryptedKeyData(
            ciphertext=unpaddedbase64.encode_base64(ciphertext),
            iv=unpaddedbase64.encode_base64(iv),
            mac=unpaddedbase64.encode_base64(digest),
        )

    def decrypt(self, event_type: str | EventType, data: EncryptedKeyData) -> bytes:
        aes_key, hmac_key = derive_keys(self.key, event_type)
        ciphertext = unpaddedbase64.decode_base64(data.ciphertext)
        mac = unpaddedbase64.decode_base64(data.mac)

        expected_mac = hmac.digest(hmac_key, ciphertext, hashlib.sha256)
        if not hmac.compare_digest(mac, expected_mac):
            raise ValueError("Invalid MAC")

        plaintext = prepare_aes(aes_key, data.iv).decrypt(ciphertext)
        return unpaddedbase64.decode_base64(plaintext.decode("utf-8"))
